require 'rspec'
require 'ruby_man/game_main'
require 'fileutils'

# Common RubyMan namespace
module RubyMan
  describe 'GameMain' do
    context 'game' do
      subject(:game) { GameMain.new }

      def select_type(arr, type)
        arr.select { |obj| obj.is_a?(type) }
      end

      it 'starts in menu' do
        expect(select_type(game.objects, Menu).size).to eq 1
      end

      it 'has at least one level' do
        expect(game.instance_variable_get(:@levels).size).to be > 0
      end

      context 'first level' do
        before { game.start }

        it 'has 3 lives' do
          expect(game.lives).to eq 3
        end

        it 'spawns correct number of required game objects' do
          expect(select_type(game.objects, Player).size).to eq 1
          expect(select_type(game.moving_objects, Player).size).to eq 1
          expect(select_type(game.objects, Food).size).to be > 0
          expect(select_type(game.objects, Hud).size).to eq 1
          # ghosts aren't mandatory for the game to work
        end

        it 'spawns player on a free spot' do
          expect(game.place_free(game.player.grid_x,
                                 game.player.grid_y)).to be true
        end

        it 'does not crash after start' do
          expect { game.update_delta(0.01) }.to_not raise_error
        end

        it 'sets player variable' do
          expect(game.player).to be_a Player
        end

        after { game.main_menu }
      end

      context 'test level' do
        def setup
          level = '|G..:' + "\n" \
                  'L_P| '
          game.end_level
          game.load_level_string(level, false)
          # verify test sanity - the player should be at (2,1)
          raise_error unless game.player.grid_x == 2 && game.player.grid_y == 1
        end

        before { setup }

        it 'player detects walls' do
          expect(game.player.direction_free(Direction.up)).to be true
          expect(game.player.direction_free(Direction.left)).to be false
          expect(game.player.direction_free(Direction.right)).to be false
        end

        context 'player is killed by ghost' do
          before do
            game.lives = 3
            ghost = game.objects_at(1, 0).first
            ghost.set_pos(2, 1)
            # process collision
            game.update_delta(0)
            # skip time after pause
            game.update_delta(2.5)
            # process level restart
            game.update_delta(0)
          end

          it 'reduces player lives' do
            expect(game.lives).to eq 2
          end

          after do
            game.lives = 3
            setup
          end
        end

        context 'ghost' do
          subject(:ghost) { game.objects_at(1, 0).first }

          before { ghost.change_direction(Direction.up) }

          it 'starts in chase mode' do
            expect(instance_variable_get(:@scared)).to be_nil
          end

          it 'knows where it can go' do
            # down would be turning back, left would be into the wall
            expect(ghost.possible_directions).to match_array [Direction.right,
                                                              Direction.up]
          end

          it 'knows directions to player' do
            expect(ghost.directions_to_player).to match_array [Direction.right,
                                                               Direction.down]
          end

          it 'knows directions in which to run away from player' do
            expect(ghost.scared_directions).to eq [Direction.up]
          end
        end

        after do
          game.main_menu
        end
      end
    end
  end
end
